import { toString } from 'hast-util-to-string';
import json5 from 'json5';
import __ from 'lodash/fp/__.js';
import findLast from 'lodash/fp/findLast.js';
import isFinite from 'lodash/fp/isFinite.js';
import kebabCase from 'lodash/fp/kebabCase.js';
import mapKeys from 'lodash/fp/mapKeys.js';
import mapValues from 'lodash/fp/mapValues.js';
import repeat from 'lodash/fp/repeat.js';
import { createHash } from 'node:crypto';
import { codeToHast } from 'shiki-renderer-hast';
import getLanguageFromCodeNode from './getLanguageFromCodeNode.js';
import isSpanElement from './lib/isSpanElement.js';
import trimNewlines from './lib/trimNewlines.js';
import parseLanguage from './parseLanguage.js';
const { parse } = json5;
const findLastFinite = findLast(isFinite);
const repeatSpaces = repeat(__, ' ');
const toDataKey = (key) => `data-${kebabCase(key)}`;
const toDataValue = (value) => (value === true ? 'true' : value);
const mapDataKeys = mapKeys(toDataKey);
const mapDataValues = mapValues(toDataValue);
const getLineNumberPadding = (maxLineNumber, lineNumber) => {
    const maxLineNumberLength = maxLineNumber.toString().length;
    const padding = lineNumber
        ? maxLineNumberLength - lineNumber.length
        : maxLineNumberLength;
    return padding > 0 ? repeatSpaces(padding) : null;
};
const isDiffSymbol = (value) => value === ' ' || value === '-' || value === '+';
export default function preNodeToShiki(highlighter, inputNode) {
    if (inputNode.tagName !== 'pre' ||
        inputNode.children.length !== 1 ||
        inputNode.children[0].type !== 'element' ||
        inputNode.children[0].tagName !== 'code' ||
        inputNode.children[0].children.length !== 1 ||
        inputNode.children[0].children[0].type !== 'text') {
        throw new Error('Expected node to be a pre element with a single child code element and text within.');
    }
    const codeNode = inputNode.children[0];
    const inputText = inputNode.children[0].children[0];
    inputText.value = trimNewlines(inputText.value);
    const meta = typeof codeNode.data?.meta === 'string' ? parse(codeNode.data.meta) : {};
    let diffSymbols = null;
    let lines = inputText.value.split('\n');
    let lineNumbers = null;
    let lineNumberPaddings = null;
    const { lang, diff } = parseLanguage(getLanguageFromCodeNode(codeNode));
    if (diff) {
        diffSymbols = lines.map((line, i) => {
            const firstChar = line.substring(0, 1);
            if (!firstChar) {
                return ' ';
            }
            if (!isDiffSymbol(firstChar)) {
                throw new Error(`First character ('${firstChar}') of line ${i + 1} is not a diff symbol. Use ' ', '-', or '+'.`);
            }
            return firstChar;
        });
        lines = lines.map(line => line.substring(1));
        inputText.value = lines.join('\n');
    }
    if (meta.lineNumbers) {
        let currentLineNumber = meta.lineNumbersOffset ?? 1;
        lineNumbers = lines.map((_, i) => meta.lineNumbers
            ? diffSymbols?.[i] === '+'
                ? null
                : (currentLineNumber++).toString()
            : null);
        const lineNumberIntegers = lineNumbers.map(lineNumber => lineNumber ? Number.parseInt(lineNumber) : null);
        const maxLineNumber = lineNumbers
            ? findLastFinite(lineNumberIntegers) ?? 1
            : 1;
        lineNumberPaddings = lines.map((_, i) => getLineNumberPadding(maxLineNumber, lineNumbers?.[i]));
    }
    const pre = codeToHast(highlighter, toString(codeNode), lang);
    const [code] = pre.children;
    if (code.type !== 'element' || code.tagName !== 'code') {
        throw new Error('Expected first child of rendered pre element to be a code element.');
    }
    const lineNodes = code.children.filter(isSpanElement);
    if (lines.length !== lineNodes.length) {
        throw new Error(`Line count did not match between text (${lines.length}) and rendered hast (${lineNodes.length}).`);
    }
    if (typeof meta.contentHash !== 'undefined') {
        const contentHash = createHash('sha1').update(inputText.value).digest('hex');
        if (contentHash !== meta.contentHash) {
            throw new Error(`Content hash mismatch. ` +
                `'${contentHash}' did not match expected hash: '${meta.contentHash}'. ` +
                `Update or remove the 'contentHash' meta property in the code block to continue.`);
        }
    }
    const properties = {
        ...mapDataKeys(mapDataValues(meta)),
        ['data-language']: lang,
    };
    pre.properties = { ...pre.properties, ...properties };
    code.properties = { ...code.properties, ...properties };
    for (const [i, n] of lineNodes.entries()) {
        const diffSymbol = diffSymbols?.[i];
        const lineNumber = lineNumbers?.[i];
        const lineNumberPadding = lineNumberPaddings?.[i];
        n.properties = {
            ...n.properties,
            ['data-line-number']: lineNumber,
            ['data-line-number-padding']: lineNumberPadding,
            dataDiffSymbol: diffSymbol,
        };
    }
    code.children = lineNodes;
    return pre;
}
